'''Copyright Tony Valier-Brasier 2024
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with this program. If not, see https://www.gnu.org/licenses/.'''

import numpy as np
import pylab as plt
import scipy.special as scp
from Functions import def_G, d_Bessel, Bessel, Hankel
#from Functions_cylsph import Bessel, Hankel


def ISA(phi,R,ki,T):
    fzero = 0
    for nn in range(np.size(T[0,:,0])):
        fzero += (2*nn + 1)*T[0,nn,:]
    Keff = 1 - 3j*phi*fzero/(ki*R)**3
    return np.sqrt(Keff)*ki


def Integral_g2_3D(r,g_r,n,ki,ks,b):
    pos = np.where(r > b)[0]
    r = r[pos]
    r_pas = r[1] - r[0]
    g_r = g_r[pos]
    
    Nf = np.size(ki)
    if np.size(ks) == 1:
        kp = ks*np.ones(Nf) 
    else:
        kp = ks
    
    r2, ki2 = np.meshgrid(r, ki)
    r2, kp2 = np.meshgrid(r, kp)
    kir = ki2*r2
    ksr = kp2*r2
    ji = scp.jv(n + 0.5, kir)*np.sqrt(np.pi/(2*kir))
    hs = scp.hankel1(n + 0.5, ksr)*np.sqrt(np.pi/(2*ksr))
    tmp = ji*hs*(g_r - 1)*kp2**3*r2**2
    tmp[0,:] = tmp[0]/2.0
    tmp[-1,:] = tmp[-1]/2.0
    I = np.sum(tmp,1)*r_pas
    return I


def MST_keff(phi,R,kL,kT,T,b,conv,corr,r=0,g2=0):
    Nf = np.size(T[0,0,:])
    Nm = np.size(T[0,:,0])
    T_LL = T[0,:,:]
    T_LT = T[1,:,:]
    T_TL = T[2,:,:]
    GLL,GTT,GTS = def_G(Nm)
    ji = Bessel(2*Nm + 1,kL*b)
    hi = Hankel(2*Nm + 1,kL*b)
    hs = Hankel(2*Nm + 1,kT*b)
    dji = d_Bessel(ji,kL*b)
    dhi = d_Bessel(hi,kL*b)
    dhs = d_Bessel(hs,kT*b)
    
    L_L = np.zeros((2*Nm,Nf),dtype=np.complex128)
    L_T = np.zeros((2*Nm,Nf),dtype=np.complex128)
    if np.size(r) > 1:
        for l in range(2*Nm):
            print('Nf : ', Nf)
            L_L[l,:] = Integral_g2_3D(r,g2,l,kL,kL,b)
            L_T[l,:] = Integral_g2_3D(r,g2,l,kL,kT,b)

    d1L = np.zeros((Nf),dtype=np.complex128)
    d2L = np.zeros((Nf),dtype=np.complex128)
    d2LL = np.zeros((Nf),dtype=np.complex128)
    d2LT = np.zeros((Nf),dtype=np.complex128)
    d2Lcor = np.zeros((Nf),dtype=np.complex128)
    d2Tcor = np.zeros((Nf),dtype=np.complex128)
    for n in range(Nm):
        d1L += (2*n + 1)*T_LL[n,:]
        for nu in range(Nm):
            Q_LL = np.zeros((Nf),dtype=np.complex128)
            N_LT = np.zeros((Nf),dtype=np.complex128)
            L_LL = np.zeros((Nf),dtype=np.complex128)
            L_LT = np.zeros((Nf),dtype=np.complex128)
            for l in range(2*Nm):
                nL = hi[l,:]*((l*(l+1) - (kL*b)**2)*ji[l,:] - kL*b*dji[l,:])
                nL += - (kL*b)**2*dji[l,:]*dhi[l,:]
                Q_LL += (-1j)**l*nL*GLL[0,n,l,nu]
                N = kL*b*dji[l,:]*hs[l,:] - kT*b*ji[l,:]*dhs[l,:]
                N_LT += (-1j)**l*N*GTT[0,n,l,nu]
                L_LL += (-1j)**l*L_L[l,:]*GLL[0,n,l,nu]
                L_LT += (-1j)**l*L_T[l,:]*GTT[0,n,l,nu]
            d2LL += 1j**(nu-n)*(2*nu + 1)*T_LL[nu,:]*T_LL[n,:]*Q_LL
            d2Lcor += 1j**(nu-n)*(2*nu + 1)*T_LL[nu,:]*T_LL[n,:]*L_LL
            if conv == 'yes':
                d2LT += 1j**(nu-n)*(2*nu + 1)*T_LT[nu,:]*T_TL[n,:]*N_LT
                if corr == 'yes':
                    d2Tcor += 1j**(nu-n)*(2*nu + 1)*T_LT[nu,:]*T_TL[n,:]*L_LT
    d2L = b/(2*R)*d2LL/(kL*R)**2 + b/R*d2LT/((kL**2 - kT**2)*R**2)
    d2L += d2Lcor/(kL*R)**3 + d2Tcor/(kT*R)**3
    Keff = 1 - 3j*phi*d1L/(kL*R)**3 - 9j*phi**2/(kL*R)**3*d2L
    return np.sqrt(Keff)*kL


def funcN(Nm,keff,b,r,gr,n0,kL,kT=0,approx_g=0):
    Nk = np.size(keff)
    NkT = np.size(kL)
    jp = np.zeros((2*Nm+1,Nk),dtype=np.complex128)
    hpL = np.zeros((2*Nm+1,NkT),dtype=np.complex128)
    hpT = np.zeros((2*Nm+1,NkT),dtype=np.complex128)
    djp = np.zeros((2*Nm,Nk),dtype=np.complex128)
    dhpL = np.zeros((2*Nm,NkT),dtype=np.complex128)
    dhpT = np.zeros((2*Nm,NkT),dtype=np.complex128)
    NL1 = np.zeros((2*Nm,Nk),dtype = np.complex128)
    NT1 = np.zeros((2*Nm,Nk),dtype = np.complex128)
    LL = np.zeros((2*Nm,Nk),dtype = np.complex128)
    LT = np.zeros((2*Nm,Nk),dtype = np.complex128)
    for n in range(2*Nm+1):
        jp[n,:] = scp.jv(n + 0.5, keff*b)*np.sqrt(np.pi/(2*keff*b))
        hpL[n,:] = scp.hankel1(n + 0.5, kL*b)*np.sqrt(np.pi/(2*kL*b))
        if kT != 0:
            hpT[n,:] = scp.hankel1(n + 0.5, kT*b)*np.sqrt(np.pi/(2*kT*b))
    for n in range(2*Nm):
        djp[n,:] = n/(keff*b)*jp[n,:] - jp[n+1,:]
        dhpL[n,:] = n/(kL*b)*hpL[n] - hpL[n+1]
        if kT != 0:
            dhpT[n,:] = n/(kT*b)*hpT[n] - hpT[n+1]
    for pp in range(2*Nm):
        NL1[pp,:] = keff*b*djp[pp,:]*hpL[pp,:] - kL*b*jp[pp,:]*dhpL[pp,:]
        NT1[pp,:] = keff*b*djp[pp,:]*hpT[pp,:] - kT*b*jp[pp,:]*dhpT[pp,:]
    
    if np.size(r) > 1:
        for pp in range(2*Nm):
            LL[pp,:] = Integral_g2_3D(r,gr,pp,keff,kL,b)
            LT[pp,:] = Integral_g2_3D(r,gr,pp,keff,kT,b)
    NL2 = 4*n0*np.pi*b*(NL1/(keff**2 - kL**2) + LL/(kL**3*b))
    NT2 = 4*n0*np.pi*b*(NT1/(keff**2 - kT**2) + LT/(kT**3*b))
    return NL2,NT2


def MatPQ_L(LL,LT,T,GLL,GTT=0):
    Nm = np.size(T[0,:])
    Nk = np.size(LL[0,:])
    PLL = np.zeros((Nm,Nm,Nk),dtype = np.complex128)
    PTT = np.zeros((Nm,Nm,Nk),dtype = np.complex128)
    QTL = np.zeros((Nm,Nm,Nk),dtype = np.complex128)
    QLT = np.zeros((Nm,Nm,Nk),dtype = np.complex128)
    for nn in range(Nm):
        for nu in range(Nm):
            for ll in range(2*Nm):
                PLL[nn,nu,:] += (-1j)**ll*LL[ll,:]*T[0,nu]*GLL[0,nn,ll,nu]
                QTL[nn,nu,:] += (-1j)**ll*LL[ll,:]*T[2,nu]*GLL[0,nn,ll,nu]
                QLT[nn,nu,:] += (-1j)**ll*LT[ll,:]*T[1,nu]*GTT[0,nn,ll,nu]
                PTT[nn,nu,:] += (-1j)**ll*LT[ll,:]*T[3,nu]*GTT[0,nn,ll,nu]
    return PLL,PTT,QTL,QLT


def dispersionL(T,NL,NT,GLL,GTT=0):
    Nm = np.size(T[0,:])
    Nk2 = np.size(NL[0,:])
    PLL,PTT,QTL,QLT = MatPQ_L(NL,NT,T,GLL,GTT)
    One = np.eye(Nm,dtype = np.complex128)
    det = np.zeros((Nk2))
    for vk in range(Nk2):
        tmp1 = np.concatenate((PLL[:,:,vk] - One, QTL[:,:,vk]), axis=1)
        tmp2 = np.concatenate((QLT[:,:,vk], PTT[:,:,vk] - One), axis=1)
        Delta = np.concatenate((tmp1,tmp2), axis=0)
        det[vk] = np.abs(np.linalg.det(Delta))    
    return det


def Dichotomie_complex_k(delta,eps,n0,R,b,r,gr,k_ref,T,kL,GLL,kT=0,GTT=0,GTS=0,plotty = 0, approx_g = 0):
    Nm = np.size(T[0,:])
    NR = 30                      # Nb points Re(solution)
    NI = 30                      # Nb points Im(solution)
    kR_min = np.real(k_ref)*(1.0 - delta)
    kR_max = np.real(k_ref)*(1.0 + delta)
    if np.imag(k_ref) == 0:
        kI_min = k_ref*10**(-6)
        kI_max = k_ref*10**(-1)
    else:
        kI_min = np.imag(k_ref)*(1.0 - delta*5)
        kI_max = np.imag(k_ref)*(1.0 + delta*5)
    
    stop = 0
    while stop < 1000:
        k_real = np.linspace(kR_min,kR_max,NR)
        k_imag = np.linspace(kI_min,kI_max,NI)
        kR_pas = k_real[1] - k_real[0]
        kI_pas = k_imag[1] - k_imag[0]
        k = np.zeros((NR,NI), dtype=np.complex128)
        for vari_r in range(NR):
            k[vari_r,:] = k_real[vari_r] + 1j*k_imag
        keff = np.ravel(k)
        NL,NT = funcN(Nm,keff,b,r,gr,n0,kL,kT,approx_g)
        det = dispersionL(T,NL,NT,GLL,GTT)

        pos = np.argmin(det)
        Amin = np.amin(det)
        Amax = np.amax(det)

        if plotty == 1:
            detty_plot = np.zeros((NR,NI))
            var = 0
            for vari_r in range(NR):
                for vari_i in range(NI):
                    detty_plot[vari_r,vari_i] = det[var]
                    var += 1
            plt.figure()
            plt.clf()
            Min_log = 10*np.log10(Amin/Amax)
            rep = 10*np.log10(np.transpose(detty_plot)/Amax)
            plt.pcolormesh(k_real,k_imag,rep,cmap="gist_heat")
            plt.clim(Min_log,Min_log+5)
            plt.colorbar()
            plt.title('STOP = ' + str(stop))
            plt.xlabel('$Re[k]$')
            plt.ylabel('$Im[k]$')
            plt.show()
            print('Root = ' + str(keff[pos]))
            print('Res = ' + str(det[pos]))
            input('Stop ???')
        if np.real(keff[pos]) == kR_min:
            kR_max = kR_min + kR_pas*5
            kR_min = kR_min*(1.0 - delta)
        elif np.real(keff[pos]) == kR_max:
            kR_min = kR_max - kR_pas*5
            kR_max = kR_max*(1.0 + delta)
        else:
            kR_min = np.real(keff[pos]) - kR_pas*5
            kR_max = np.real(keff[pos]) + kR_pas*5
        if np.imag(keff[pos]) == kI_min:
            kI_max = kI_min + kI_pas
            kI_min = kI_min*(1.0 - delta)
        elif np.imag(keff[pos]) == kI_max:
            kI_min = kI_max - kI_pas
            kI_max = kI_max*(1.0 + delta)
        else:
            kI_min = np.imag(keff[pos]) - kI_pas*5
            kI_max = np.imag(keff[pos]) + kI_pas*5
        if kI_min < 0:
            kI_min = 0
        stop += 1
        if (kR_pas <= eps and kI_pas <= eps):
            stop = 1000
    K = (kR_min + kR_max)/2.0 + 1j*(kI_min + kI_max)/2.0
    return K


def MST_exact(phi,R,kL,kT,T,b,r=0,g2=0,plotty=0):
    ''' This code solves numerically the dispersion equation. That can be optimsed.
    In some configurations, this code takes a long time to determine the solution.
    A solution is to modify the parameter delta.'''
    n0 = phi/(4/3*np.pi*R**3)
    Nm = np.size(T[0,:,0])
    Nf = np.size(T[0,0,:])
    GLL,GTT,GTS = def_G(Nm)
    k1 = MST_keff(phi,R,kL,kT,T,b,'yes','yes')
    K = np.zeros((Nf),dtype = np.complex128)
    eps = 1e-1
    for vf in range(Nf):
        print(vf)
        if vf <= 3:
            kref = k1[vf]
            delta = 0.05
        else:
            kref = 2*K[vf-1] - K[vf-2]
            delta = 0.001
        K[vf] = Dichotomie_complex_k(delta,eps,n0,R,b,r,g2,kref,T[:,:,vf],kL[vf],GLL,kT[vf],GTT,GTS,plotty)
    return K

