'''Copyright Tony Valier-Brasier 2024
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with this program. If not, see https://www.gnu.org/licenses/.'''

import numpy as np
import scipy.special as scp


def Solide_Sphere_elastique(Nm, f, R, lam_p, mu_p, rho_p, lam_m, mu_m, rho_m):
    Nf = np.size(f)
    Tn = np.zeros((4,Nm,Nf),dtype=np.complex128)
    MU = mu_p/mu_m*np.ones(Nf,dtype=np.complex128)
    K_EL = 2*np.pi*f*R*np.sqrt(rho_m/(lam_m + 2*mu_m))
    K_ET = 2*np.pi*f*R*np.sqrt(rho_m/mu_m)
    K_IL = 2*np.pi*f*R*np.sqrt(rho_p/(lam_p + 2*mu_p))
    K_IT = 2*np.pi*f*R*np.sqrt(rho_p/mu_p)

    for vf in range(Nf):
        mu = MU[vf]
        J_SL = np.zeros((Nm+1),dtype=np.complex128)
        J_ST = np.zeros((Nm+1),dtype=np.complex128)
        J_EL = np.zeros((Nm+1),dtype=np.complex128)
        J_ET = np.zeros((Nm+1),dtype=np.complex128)
        H_EL = np.zeros((Nm+1),dtype=np.complex128)
        H_ET = np.zeros((Nm+1),dtype=np.complex128)
        for nu in range(Nm+1):
            J_SL[nu] = scp.jn(nu + 0.5, K_IL[vf])*np.sqrt(np.pi/(2*K_IL[vf]))
            J_ST[nu] = scp.jn(nu + 0.5, K_IT[vf])*np.sqrt(np.pi/(2*K_IT[vf]))
            J_EL[nu] = scp.jn(nu + 0.5, K_EL[vf])*np.sqrt(np.pi/(2*K_EL[vf]))
            J_ET[nu] = scp.jn(nu + 0.5, K_ET[vf])*np.sqrt(np.pi/(2*K_ET[vf]))
        H_EL = Hankel(Nm+1,K_EL[vf])
        H_ET = Hankel(Nm+1,K_ET[vf])
        Z_ext = Vector(0, 1, K_EL[vf], K_ET[vf], H_EL, H_ET)
        Z_inc = Vector(0, 1, K_EL[vf], K_ET[vf], J_EL, J_ET)
        Z_int = Vector(0, mu, K_IL[vf], K_IT[vf], J_SL, J_ST)
        
        D = np.zeros((4),dtype=np.complex128)
        F = np.zeros((2),dtype=np.complex128)
        for pp in range(2):
            D[pp*2] = Z_ext[pp*2]
            D[pp*2 + 1] = - Z_int[pp*2]
            F[pp] = Z_inc[pp*2]
        Tn[0,0,vf] = - (F[0]*D[3] - D[1]*F[1])/(D[0]*D[3] - D[1]*D[2])*K_EL[vf]*np.exp(- 1j*K_EL[vf])

        for n in range(1,Nm):
            Z_ext = Vector(n, 1, K_EL[vf], K_ET[vf], H_EL, H_ET)
            Z_inc = Vector(n, 1, K_EL[vf], K_ET[vf], J_EL, J_ET)
            Z_int = Vector(n, mu, K_IL[vf], K_IT[vf], J_SL, J_ST)
            DI = np.zeros((16),dtype=np.complex128)
            DLL = np.zeros((16),dtype=np.complex128)
            DLT = np.zeros((16),dtype=np.complex128)
            DTL = np.zeros((16),dtype=np.complex128)
            DTT = np.zeros((16),dtype=np.complex128)
            for pp in range(4):
                DI[pp*4] = Z_ext[pp]
                DI[pp*4 + 1] = Z_ext[pp+4]
                DI[pp*4 + 2] = - Z_int[pp]
                DI[pp*4 + 3] = - Z_int[pp+4]

                DLL[pp*4] = Z_inc[pp]
                DLL[pp*4 + 1] = Z_ext[pp+4]
                DLL[pp*4 + 2] = - Z_int[pp]
                DLL[pp*4 + 3] = - Z_int[pp+4]

                DLT[pp*4] = Z_ext[pp]
                DLT[pp*4 + 1] = Z_inc[pp]
                DLT[pp*4 + 2] = - Z_int[pp]
                DLT[pp*4 + 3] = - Z_int[pp+4]

                DTL[pp*4] = Z_inc[pp+4]
                DTL[pp*4 + 1] = Z_ext[pp+4]
                DTL[pp*4 + 2] = - Z_int[pp]
                DTL[pp*4 + 3] = - Z_int[pp+4]

                DTT[pp*4] = Z_ext[pp]
                DTT[pp*4 + 1] = Z_inc[pp+4]
                DTT[pp*4 + 2] = - Z_int[pp]
                DTT[pp*4 + 3] = - Z_int[pp+4]

            det = det_comp4(DI)
            Tn[0,n,vf] = - det_comp4(DLL)/det*K_EL[vf]*np.exp(- 1j*K_EL[vf])
            Tn[1,n,vf] = - det_comp4(DLT)/det*K_ET[vf]*np.exp(- 1j*K_ET[vf])
            Tn[2,n,vf] = - det_comp4(DTL)/det*K_EL[vf]*np.exp(- 1j*K_EL[vf])
            Tn[3,n,vf] = - det_comp4(DTT)/det*K_ET[vf]*np.exp(- 1j*K_ET[vf])
    return Tn


def Vector(n, mu, KL, KT, ZnL, ZnT):
    Zeta = np.zeros((8),dtype=np.complex128)
    Zeta[0] = n*ZnL[n] - KL*ZnL[n+1]
    Zeta[1] = ZnL[n]
    Zeta[2] = mu*((2*(n-1)*n - KT**2)*ZnL[n] + 4*KL*ZnL[n+1])
    Zeta[3] = mu*2*((n-1)*ZnL[n] - KL*ZnL[n+1])

    Zeta[4] = (n+1)*n*ZnT[n]
    Zeta[5] = (n+1)*ZnT[n] - KT*ZnT[n+1]
    Zeta[6] = 2*n*(n+1)*mu*((n-1)*ZnT[n] - KT*ZnT[n+1])
    Zeta[7] = mu*((2*(n**2-1) - KT**2)*ZnT[n] + 2.0*KT*ZnT[n+1])
    return Zeta

def det_comp2(D):
    res = D[0]*D[3] - D[1]*D[2]
    return res;

def det_comp4(D):
    res = D[0]*(D[5]*(D[10]*D[15] - D[11]*D[14]))
    res += - D[0]*(D[9]*(D[6]*D[15] - D[7]*D[14]))
    res += D[0]*(D[13]*(D[6]*D[11] - D[7]*D[10]))

    res += - D[4]*(D[1]*(D[10]*D[15] - D[11]*D[14]))
    res += D[4]*(D[9]*(D[2]*D[15] - D[3]*D[14]))
    res += - D[4]*(D[13]*(D[2]*D[11] - D[3]*D[10]))

    res += D[8]*(D[1]*(D[6]*D[15] - D[7]*D[14]))
    res += - D[8]*(D[5]*(D[2]*D[15] - D[3]*D[14]))
    res += D[8]*(D[13]*(D[2]*D[7] - D[3]*D[6]))

    res += - D[12]*(D[1]*(D[6]*D[11] - D[7]*D[10]))
    res += D[12]*(D[5]*(D[2]*D[11] - D[3]*D[10]))
    res += - D[12]*(D[9]*(D[2]*D[7] - D[3]*D[6]))
    return res

def Hankel(n,x):
    H = np.zeros((n,np.size(x)),dtype=np.complex128)
    H[0,:] = - 1j
    if n == 2:
        H[1,:] = - 1 - 1j/x
    elif n == 3:
        H[1,:] = - 1 - 1j/x
        H[2,:] = 1j - 3/x - 3j/x**2
    else:
        H[1,:] = - 1 - 1j/x
        H[2,:] = 1j - 3/x - 3j/x**2
        p = 2
        while p < (n-1):
            p = p + 1
            H[p,:] = (2*p - 1)*H[p - 1,:]/x - H[p - 2,:]
    return H
