function fac=thermMultFac(ai,aj,K,L)
%These are the terms in the multiple scattering formula additional term 
%for thermal multiple scattering derived by Luppe (JASA 2012) and VJP Dec
%2012-Jan 2013 and July 2013
%It calculates the terms:
%-Lb*(Kb*j0pKb*h0(Lb)-Lb*j0(Kb)*h0p(Lb))*j0(La)/h0(La)
%This is designed for VJP version which uses scaled scattering
%coefficients
%where b=ai+aj (sum of the two particle radii)
%and the factor is taken as the average using ai and aj as a
%ai and aj are the radius of the two particles (metres)
%
b=ai+aj;%the exclusion zone radius
Lb=L*b; 
Kb=K*b; %b=the exclusion zone radius
%calculate the required Bessel/Hankel functions
h0Lb=-1i*exp(1i*Lb)/Lb;
j0Kb=sin(Kb)/Kb;
j0pKb=((1/tan(Kb))-1/Kb)*j0Kb;
Lai=L*ai; %this is=La for the first particle, radius ai
Laj=L*aj; %this is La for the second particle, radius aj
j0Lai_h0Laj=0.5*(aj/ai)*(exp(1i*(Lai-Laj))-exp(-1i*(Lai+Laj)));
j0Laj_h0Lai=0.5*(ai/aj)*(exp(1i*(Lai-Laj))-exp(-1i*(Lai+Laj)));

facb=((1i*Lb-1)*j0Kb-Kb*j0pKb)*Lb*h0Lb; %these are the terms which depend only on b

faci=facb*j0Laj_h0Lai;
facj=facb*j0Lai_h0Laj;

%take the average factor
fac=0.5*(faci+facj);
